﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Xenon.KifuNarabe.L01_Communication;

namespace Xenon.KifuNarabe.L02_Common
{

    /// <summary>
    /// ************************************************************************************************************************
    /// 升目の場所、または　駒がどこからどこへ動いたかの一手、を表すデータです。
    /// ************************************************************************************************************************
    /// 
    /// 盤上用の升目符号。「元」→「現」の形。
    /// 
    /// 「先後」、
    /// 「置き場」、
    /// 「筋、段」（明記がない場合、dstの意）、
    /// 「src筋、src段」（オプション）。
    /// 
    /// ・“同”
    /// 棋譜用の升目符号。“同”が使えます。
    /// “同”にする場合は、UNKNOWN_SUJI 筋、UNKNOWN_DAN 段に設定してください。毎回再計算します。
    /// 
    /// ・前の升：「7g7f」などにも使えるよう、src筋、src段も補助で用意。
    /// ・前の手：また、遡れるように、１つ前の升目符号へのリンクも補助で容易。
    /// 
    /// ・駒種類：「歩」「と金」など。補助で容易。
    /// </summary>
    [Serializable]
    public class TeProcess
    {

        #region 定数

        /// <summary>
        /// 働かない値として、筋に埋めておくためのものです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int CTRL_NOTHING_PROPERTY_SUJI = int.MinValue;

        /// <summary>
        /// 働かない値として、段に埋めておくためのものです。8～-8 程度だと、角等の射程に入るので、大きく外した数字をフラグに使います。
        /// </summary>
        public const int CTRL_NOTHING_PROPERTY_DAN = int.MinValue;

        //------------------------------------------------------------
        /// <summary>
        /// 筋は 1～9 だけ有効です。
        /// </summary>
        public const int YUKO_SUJI_MIN = 1;
        public const int YUKO_SUJI_MAX = 9;

        /// <summary>
        /// 段は 1～9 だけ有効です。
        /// </summary>
        public const int YUKO_DAN_MIN = 1;
        public const int YUKO_DAN_MAX = 9;
        //------------------------------------------------------------
        #endregion



        #region プロパティー類

        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 先後
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public Sengo Sengo
        {
            get
            {
                return this.sengo;
            }
        }
        private Sengo sengo;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 現・駒の置き場
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public Okiba Okiba
        {
            get
            {
                return this.okiba;
            }
        }
        private Okiba okiba;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 元・駒の置き場
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public Okiba SrcOkiba
        {
            get
            {
                return this.srcOkiba;
            }
        }
        private Okiba srcOkiba;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 現・筋。
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public virtual int Suji
        {
            get
            {
                return this.suji;
            }
        }
        private int suji;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 元・筋
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public virtual int SrcSuji
        {
            get
            {
                return this.srcSuji;
            }
        }
        private int srcSuji;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 現・段。
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public int Dan
        {
            get
            {
                return this.dan;
            }
        }
        private int dan;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 元・段
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public virtual int SrcDan
        {
            get
            {
                return this.srcDan;
            }
        }
        private int srcDan;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 現・駒の種類。歩、桂…。
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public KomaSyurui Syurui
        {
            get
            {
                return this.syurui;
            }
        }
        private KomaSyurui syurui;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// 元・駒の種類。歩、桂…。  ＜不成判定用＞
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public KomaSyurui SrcSyurui
        {
            get
            {
                return this.srcSyurui;
            }
        }
        private KomaSyurui srcSyurui;


        /// <summary>
        /// ------------------------------------------------------------------------------------------------------------------------
        /// あれば、取った駒。
        /// ------------------------------------------------------------------------------------------------------------------------
        /// </summary>
        public KomaSyurui TottaKoma
        {
            get
            {
                return this.tottaKoma;
            }
        }
        private KomaSyurui tottaKoma;

        #endregion


        /// <summary>
        /// ************************************************************************************************************************
        /// 駒用。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="sengo"></param>
        /// <param name="okiba"></param>
        /// <param name="srcOkiba"></param>
        /// <param name="suji"></param>
        /// <param name="dan"></param>
        /// <param name="srcSuji"></param>
        /// <param name="srcDan"></param>
        /// <param name="syurui"></param>
        /// <param name="srcSyurui"></param>
        /// <param name="previousTe"></param>
        public TeProcess(Sengo sengo, Okiba okiba, int suji, int dan, KomaSyurui syurui)
        {
            this.sengo = sengo;
            this.okiba = okiba;
            this.srcOkiba = Okiba.CTRL_NOTHING_PROPERTY;
            this.suji = suji;
            this.dan = dan;
            this.srcSuji = TeProcess.CTRL_NOTHING_PROPERTY_SUJI;
            this.srcDan = TeProcess.CTRL_NOTHING_PROPERTY_DAN;
            this.syurui = syurui;
            this.srcSyurui = KomaSyurui.CTRL_NOTHING_PROPERTY;
            this.tottaKoma = KomaSyurui.CTRL_NOTHING_PROPERTY;
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 棋譜用。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="sengo"></param>
        /// <param name="okiba"></param>
        /// <param name="srcOkiba"></param>
        /// <param name="suji"></param>
        /// <param name="dan"></param>
        /// <param name="srcSuji"></param>
        /// <param name="srcDan"></param>
        /// <param name="syurui"></param>
        /// <param name="srcSyurui"></param>
        /// <param name="previousTe"></param>
        public TeProcess(Sengo sengo, Okiba okiba, Okiba srcOkiba, int suji, int dan, int srcSuji, int srcDan, KomaSyurui syurui, KomaSyurui srcSyurui, KomaSyurui tottaKoma)
        {
            this.sengo = sengo;
            this.okiba = okiba;
            this.srcOkiba = srcOkiba;
            this.suji = suji;
            this.dan = dan;
            this.srcSuji = srcSuji;
            this.srcDan = srcDan;
            this.srcSyurui = srcSyurui;
            this.syurui = syurui;
            this.tottaKoma = tottaKoma;
        }


        /// <summary>
        /// ************************************************************************************************************************
        /// 元位置。
        /// ************************************************************************************************************************
        /// </summary>
        /// <returns></returns>
        public TeProcess Src()
        {
            TeProcess src = new TeProcess(
                this.Sengo,
                this.Okiba,
                this.SrcOkiba,
                this.srcSuji,
                this.srcDan,
                TeProcess.CTRL_NOTHING_PROPERTY_SUJI,
                TeProcess.CTRL_NOTHING_PROPERTY_DAN,
                this.SrcSyurui,
                KomaSyurui.CTRL_NOTHING_PROPERTY,
                KomaSyurui.CTRL_NOTHING_PROPERTY
                );

            return src;
        }


        #region 比較判定系

        /// <summary>
        /// ************************************************************************************************************************
        /// 置き場一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="okiba2"></param>
        /// <returns></returns>
        public bool MatchOkiba(Okiba okiba2)
        {
            return
                this.okiba == okiba2;
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 置き場一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="m2"></param>
        /// <returns></returns>
        public bool MatchOkiba(TeProcess m2)
        {
            return this.MatchOkiba(m2.Okiba);
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 先後一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="m2"></param>
        /// <returns></returns>
        public bool MatchSengo(TeProcess m2)
        {
            return
                this.sengo == m2.sengo;
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 筋一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="suji2"></param>
        /// <returns></returns>
        public bool MatchSuji(int suji2)
        {
            return
                // 有効な値かどうか
                TeProcess.YUKO_SUJI_MIN<=this.suji
                && this.suji<=TeProcess.YUKO_SUJI_MAX

                && this.suji == suji2;
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 筋一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="m2"></param>
        /// <returns></returns>
        public bool MatchSuji(TeProcess m2)
        {
            return this.MatchSuji(m2.Suji);
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 段一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="dan2"></param>
        /// <returns></returns>
        public bool MatchDan(int dan2)
        {
            return
                // 有効な値かどうか
                TeProcess.YUKO_DAN_MIN <= this.dan
                && this.dan <= TeProcess.YUKO_DAN_MAX

                && this.dan == dan2;
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 段一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="m2"></param>
        /// <returns></returns>
        public bool MatchDan(TeProcess m2)
        {
            return this.MatchDan(m2.Dan);
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 升一致判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="obj2"></param>
        /// <returns></returns>
        public bool MatchMasu(Object obj2)
        {
            bool result = false;

            if(obj2 is TeProcess)
            {
                TeProcess m2 = (TeProcess)obj2;
                result = this.MatchSuji(m2) && this.MatchDan(m2);
            }

            return result;
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 一致判定：   升目符号同士
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="masu2"></param>
        /// <returns></returns>
        public bool Match(TeProcess masu2)
        {
            return
                this.MatchOkiba(masu2)
                && this.MatchSuji(masu2)
                && this.MatchDan(masu2);
        }

        /// <summary>
        /// ************************************************************************************************************************
        /// 含まれるか判定。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="masu2Arr"></param>
        /// <returns></returns>
        public bool ExistsIn(List<TeProcess> masu2Arr)
        {
            bool matched = false;

            foreach (TeProcess m in masu2Arr)
            {
                if (
                    null != m
                    && this.MatchSengo(m) //2014-07-21 先後も見るように追記。
                    && this.MatchSuji(m)
                    && this.MatchDan(m)
                    )
                {
                    matched = true;
                    break;
                }
            }

            return matched;
        }

        #endregion



        /// <summary>
        /// ************************************************************************************************************************
        /// １マス上、のように指定して、マスを取得します。
        /// ************************************************************************************************************************
        /// </summary>
        /// <param name="offsetSuji"></param>
        /// <param name="offsetDan"></param>
        /// <returns></returns>
        public TeProcess Offset(Sengo sengo, int offsetSuji, int offsetDan)
        {
            return new TeProcess(
                sengo, 
                this.Okiba,
                this.SrcOkiba,
                this.Suji + offsetSuji,
                this.Dan + offsetDan,
                this.SrcSuji,
                this.SrcDan,
                this.Syurui,
                this.SrcSyurui,
                this.TottaKoma
                );
        }


        /// <summary>
        /// ************************************************************************************************************************
        /// SFEN符号表記。
        /// ************************************************************************************************************************
        /// </summary>
        /// <returns></returns>
        public string ToSfenText()
        {
            StringBuilder sb = new StringBuilder();


            if (this.IsDaAction)
            {
                // 打でした。
                //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                // (自)筋・(自)段は書かずに、「P*」といった表記で埋めます。
                sb.Append(Converter01.SyuruiToSfenDa(this.SrcSyurui));
                sb.Append("*");
            }
            else
            {
                //------------------------------------------------------------
                // (自)筋
                //------------------------------------------------------------
                string strSrcSuji;
                switch (this.SrcSuji)
                {
                    case CTRL_NOTHING_PROPERTY_SUJI:
                        strSrcSuji = "Ｎ筋";//エラー表現
                        break;
                    default:
                        strSrcSuji = this.SrcSuji.ToString();
                        break;
                }
                sb.Append(strSrcSuji);

                //------------------------------------------------------------
                // (自)段
                //------------------------------------------------------------
                string strSrcDan;
                switch (this.SrcDan)
                {
                    case CTRL_NOTHING_PROPERTY_DAN:
                        strSrcDan = "Ｎ段";//エラー表現
                        break;
                    default:
                        strSrcDan = Converter01.IntToAlphabet(this.SrcDan);
                        break;
                }
                sb.Append(strSrcDan);
            }

            //------------------------------------------------------------
            // (至)筋
            //------------------------------------------------------------
            string strSuji;
            switch (this.Suji)
            {
                case CTRL_NOTHING_PROPERTY_SUJI:
                    strSuji = "Ｎ筋";//エラー表現
                    break;
                default:
                    strSuji = this.Suji.ToString();
                    break;
            }
            sb.Append(strSuji);


            //------------------------------------------------------------
            // (至)段
            //------------------------------------------------------------
            string strDan;
            switch (this.Dan)
            {
                case CTRL_NOTHING_PROPERTY_DAN:
                    strDan = "Ｎ段";//エラー表現
                    break;
                default:
                    strDan = Converter01.IntToAlphabet(this.Dan);
                    break;
            }
            sb.Append(strDan);


            //------------------------------------------------------------
            // 成
            //------------------------------------------------------------
            if (this.IsNatta_Process)
            {
                sb.Append("+");
            }

            return sb.ToString();
        }

        public override string ToString()
        {
            return this.ToSfenText();
        }


        /// <summary>
        /// 移動前と、移動後の場所が異なっていれば真。
        /// </summary>
        /// <returns></returns>
        public bool DoneMove()
        {
            return !this.Match(this.Src());
        }

        /// <summary>
        /// 不成ケース
        /// </summary>
        /// <returns></returns>
        public KomaSyurui ToFunariCase()
        {
            return Converter01.SyuruiToFunariCase(this.Syurui);
        }

        /// <summary>
        /// 成ケース
        /// </summary>
        /// <returns></returns>
        public KomaSyurui ToNariCase()
        {
            return Converter01.SyuruiToNariCase(this.Syurui);
        }

        /// <summary>
        /// 駒の表面の文字。
        /// </summary>
        public string Text_Label
        {
            get
            {
                return Converter01.SyuruiToLabel(this.Syurui);
            }
        }

        /// <summary>
        /// 成り
        /// </summary>
        public bool Nari
        {
            get
            {
                return Converter01.SyuruiToNari(this.Syurui);
            }
        }

        /// <summary>
        /// 成った
        /// </summary>
        /// <returns></returns>
        public bool IsNatta_Process
        {
            get
            {
                return !Converter01.IsSyuruiNari(this.SrcSyurui) && Converter01.IsSyuruiNari(this.Syurui);
            }
        }

        /// <summary>
        /// “打” ＜アクション時＞
        /// </summary>
        /// <returns></returns>
        public bool IsDaAction
        {
            get
            {
                return
                    Okiba.ShogiBan != this.SrcOkiba
                    && Okiba.CTRL_NOTHING_PROPERTY != this.SrcOkiba;//初期配置から移動しても、打にはしません。
            }
        }

        /// <summary>
        /// 不一致判定：　先後、駒種類  が、自分と同じものが　＜ひとつもない＞
        /// </summary>
        /// <returns></returns>
        public bool NeverOnaji(Kyokumen kyokumen, params List<int>[] komaHandleGroupArgs)
        {
            bool unmatched = true;

            foreach (List<int> komaHandleGroup in komaHandleGroupArgs)
            {
                foreach (int komaHandle in komaHandleGroup)
                {
                    TeProcess koma = kyokumen.KomaDoors[komaHandle];

                    if (
                            this.Sengo == koma.Sengo
                        && this.Syurui == koma.Syurui
                        )
                    {
                        // １つでも一致するものがあれば、終了します。
                        unmatched = false;
                        goto gt_EndLoop;
                    }
                }

            }
        gt_EndLoop:

            return unmatched;
        }

        /// <summary>
        /// 相手陣に入っていれば真。
        /// </summary>
        /// <returns></returns>
        public bool InAitejin()
        {
            return
                // 後手は 7,8,9 段。
                (Sengo.Gote == this.Sengo && 7 <= this.Dan)
                // 先手は 1,2,3 段。
                || (Sengo.Sente == this.Sengo && this.Dan <= 3);
        }

    }

}
